
#import <Accelerate/Accelerate.h>
#import "GradientController.h"

static const float	gradientDomain[] = { 0.0, 1.0 };
static const float	colorRanges[] = { 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0 };

typedef struct ColorVector
{
	float	r;
	float	g;
	float	b;
	float	a;
} ColorVector;

static void evaluateLinearGradient(void *info, const float *in, float *out)
{
	ColorVector		*colors;
	
	colors = info;
	
	*out++ = colors[0].r + *in * colors[1].r;
	*out++ = colors[0].g + *in * colors[1].g;
	*out++ = colors[0].b + *in * colors[1].b;
	*out++ = colors[0].a + *in * colors[1].a;
}

static const CGFunctionCallbacks scalarCallbacks = { 0, &evaluateLinearGradient, NULL };

typedef union
{
	float			scalar[4];
	vFloat			vector;
} AlignedVector;

#ifdef __ALTIVEC__
static void evaluateLinearGradientAltivec(void *info, const float *in, float *out)
{
	vFloat			*colors;
	AlignedVector	result, multiplier;
	
	colors = (vFloat *)info;
	
	multiplier.scalar[0] = *in;
	multiplier.scalar[1] = *in;	
	multiplier.scalar[2] = *in;
	multiplier.scalar[3] = *in;

	result.vector = vec_madd(multiplier.vector, colors[1], colors[0]);

	vec_st(result.vector, 0, out);
}

static const CGFunctionCallbacks vectorCallbacks = { 0, &evaluateLinearGradientAltivec, NULL };

#endif

// My guess at what the SSE version would look like. I don't have an Intel machine to test this
// or enough knowledge of SSE to even know if I'm close.
/*
static void evaluateLinearGradientSSE(void *info, const float *in, float *out)
{
	vFloat			*colors;
	AlignedVector	multiplier;
	AlignedVector	result;
	
	colors = (vFloat *)info;
	
	multiplier.scalar[0] = *in;
	multiplier.scalar[1] = *in;	
	multiplier.scalar[2] = *in;
	multiplier.scalar[3] = *in;
	
	result.vector = _mm_add_ps(_mm_mul_ps(multiplier.vector, colors[1]), colors[0]);
	
	//	*out++ = result.scalar[0];
	//	*out++ = result.scalar[1];
	//	*out++ = result.scalar[2];
	//	*out++ = result.scalar[3];
	_mm_store_ps(out, result.vector);
}*/


@implementation GradientController

- (void)drawGradientInRect:(NSRect)bounds vectorize:(BOOL)flag
{
	CGColorSpaceRef		colorspace;
	CGFunctionRef		function;
	CGShadingRef		shading;
	CGContextRef		context;
	AlignedVector		vector;
	vFloat				colorVectors[2];
	ColorVector			colors[2];
	CGPoint				startPoint, endPoint;
		
	colorspace = CGColorSpaceCreateDeviceRGB();
	context = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
	function = NULL;
	
	if (flag)
	{
		// Black
		vector.scalar[0] = 0.0;
		vector.scalar[1] = 0.0;
		vector.scalar[2] = 0.0;
		vector.scalar[3] = 1.0;
		
		colorVectors[0] = vector.vector;
		
		// Store the difference between white and black, not the absolute end color. Saves a subtraction in the gradient calculation
		vector.scalar[0] = 1.0;
		vector.scalar[1] = 1.0;
		vector.scalar[2] = 1.0;
		vector.scalar[3] = 0.0;
		
		colorVectors[1] = vector.vector;
		
#ifdef __ALTIVEC__
		function = CGFunctionCreate(colorVectors, 1, gradientDomain, 4, colorRanges, &vectorCallbacks);
#endif
	}
	else
	{
		colors[0].r = 0.0;
		colors[0].g = 0.0;
		colors[0].b = 0.0;
		colors[0].a = 1.0;
		
		// Store the difference between white and black, not the absolute end color. Saves a subtraction in the gradient calculation
		colors[1].r = 1.0;
		colors[1].g = 1.0;
		colors[1].b = 1.0;
		colors[1].a = 0.0;
		
		function = CGFunctionCreate(colors, 1, gradientDomain, 4, colorRanges, &scalarCallbacks);
	}
	
	if (function == NULL)
	{
		//PENDING:
		NSLog(@"Could not create gradient function");
		return;
	}
	
	startPoint.x = NSMinX(bounds);
	startPoint.y = NSMinY(bounds);
	endPoint.x = NSMaxX(bounds);
	endPoint.y = startPoint.y;
	
	shading = CGShadingCreateAxial(colorspace, *(CGPoint *)&startPoint, *(CGPoint *)&endPoint, function, false, false);
	
	CGContextSaveGState(context);
	
	CGContextDrawShading(context, shading);
	CGContextRestoreGState(context);
	
	CGFunctionRelease(function);	
	CGShadingRelease(shading);
	CGColorSpaceRelease(colorspace);
}

- (IBAction)run:(id)sender
{
	NSColor			*color1, *color2;
	NSTimeInterval	startTime, endTime;
	int				iterations, i;
	BOOL			vectorize;
	
	color1 = [NSColor blackColor];
	color2 = [NSColor whiteColor];
	iterations = [iterationsField intValue];

	vectorize = [methodPopUp selectedTag] == 1;
	[gradientView lockFocus];
	
	startTime = [NSDate timeIntervalSinceReferenceDate];
	
	for (i = 0; i < iterations; i++)
	{
		[self drawGradientInRect:[gradientView bounds] vectorize:vectorize];
	}
	
	endTime = [NSDate timeIntervalSinceReferenceDate];
	
	[gradientView unlockFocus];
	
	[timeField setDoubleValue:(endTime - startTime)];
	[averageTimeField setDoubleValue:(endTime - startTime) / iterations];
}

@end
